/******************************************************************************
 *   "$Id:  $"
 *
 *                 Copyright (c) 2000  O'ksi'D
 *
 *                      All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *      Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *      Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *      Neither the name of O'ksi'D nor the names of its contributors
 *      may be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER 
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *   Author : Jean-Marc Lienher ( http://oksid.ch )
 *
 ******************************************************************************/


#include "xd640/Xd6Std.h"
#include "gui.h"
//#include "flwriter.xpm"
#include "xd640/Xd6IconWindow.h"
//#include "Fl_Font.H"
#include <iostream>
using namespace std;

#if MSDOS
#include <sys/exceptn.h>
#endif

#  ifdef MSDOS
#     include <mwin/nano-X.h>
#	  include <IO.h>
extern "C" {
void toggle_text_mode();
}
#  endif

#include <string>
using std::string;

#define PACKAGE "flwriter"
#define LOCALEDIR "/usr/local/share/locale"

Xd6ConfigFile *cfg;

int fl_font_ =0;
int fl_size_ =12;

int main(int argc, char **argv, char **environ)
{
	int index = 0;
	GUI *gui = NULL;
	Xd6ConfigFileSection *cfg_sec = NULL;
	Xd6ConfigFileItem *itm = NULL;
	const char *val = NULL;
	Fl_Font font = (Fl_Font) 0;
	int size = 16;
	char title_cstr_m[100];


#if MSDOS
__djgpp_set_ctrl_c(0);
//select the grad1 scheme - looks a bit like Win7
setenv("FLTK_SCHEME","grad1",0); //do not replace existing
#endif

	Fl::args(argc, argv, index);

	fl_open_display();

#if MSDOS
/* This is apparently not needed but left in for safety. In case memory is
   used uninitialised and assumed to be zero, this here will clear 2 MB at
   program start which will be used later again but then set to zero.
*/
char tmem[100];
char* tempmem = (char*)calloc(20000,sizeof(tmem)); //alloc 2 MB
if (tempmem != NULL) {
	cfree(tempmem);} 
else { //could not allocate 2 MB - try 750k
	tempmem = (char*)calloc(7500,sizeof(tmem)); //alloc 750 k
	if (tempmem != NULL) cfree(tempmem);
}
//GrClose();
//exit(0);
#endif

	// initialize an FLE configuration object
	cfg = new Xd6ConfigFile("flwriter", "Utilities");
	cfg_sec = cfg->get_config_section(cfg->app_name);

#if 0 //#ifndef MSDOS
	// GNU gettext initialization
	setlocale(LC_MESSAGES, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);
#endif

	// load FLE default fonts	
	Xd6DefaultFonts::load(cfg);
        if (cfg_sec) {
                itm = cfg_sec->get_item("FontId", NULL);
                val = NULL; if (itm) val = itm->get_value();
                if (val) font = (Fl_Font) strtoul(val, NULL, 0);
                itm = cfg_sec->get_item("FontSize", NULL);
                val = NULL; if (itm) val = itm->get_value();
                if (val) size = (int) strtoul(val, NULL, 0);
        }
	if ((int)font >= (int) Xd6DefaultFonts::free_font) font = (Fl_Font) 0;

	// create the user interface
   	gui = new GUI(240, 190);
	gui->font = font;
	gui->size = size;
	gui->cfg = cfg;
	gui->cfg_sec = cfg_sec;
   	gui->create_layout();

	// show the interface
	gui->resize(gui->x(), gui->y(), 620, 440);
//toggle_text_mode();
#if MSDOS
        if (argc>1){
        if (strcmp(argv[index],"-w")!=0) gui->fullscreen(); //fullscreen is default
        } else {
        gui->fullscreen(); //fullscreen is default
        }
#endif
	if (index != argc) {
	if (argc>2) index++;
	if (strcmp(argv[index],"-w")!=0){ // wrongly -w as second parameter?
			gui->load(argv[index]);
			//titlestring = "FlWriter 2012 - Currently editing: ";
			//titlestring += GUI::self->editor->file;
			//GUI::self->label(titlestring.c_str());
			sprintf(title_cstr_m,"FlWriter 2012 - Currently editing: %s",GUI::self->editor->file);
			GUI::self->copy_label(title_cstr_m);
	} //strcmp
	} //!=argc
	
	gui->show(argc, argv);
	
	//set_wm_icon(flwriter_xpm, gui); //removed - gp
	// main loop
   	return Fl::run();
}

